/*
 *      lark-rhvoice.c
 *
 *      Copyright 2016-2018 Alex <alex@linuxonly.ru>
 *
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <RHVoice.h>
#include <RHVoice_common.h>
#include <string.h>

#include "lark-rhvoice.h"
#include "lark-gstreamer.h"

#define LARK_RHVOICE_FORMAT      LARK_GSTREAMER_SRC_FORMAT_S16LE
#define LARK_RHVOICE_SAMPLERATE  24000
#define LARK_RHVOICE_CHANNELS    1

typedef int (*rhvoice_play_speech_func)(const short *samples, unsigned int count, void *user_data);

struct _LarkRHVoice
{
	GObject parent_instance;
	/*Data*/
	RHVoice_tts_engine engine;
	gboolean *genstate;
};

struct _LarkRHVoiceClass
{
	GObjectClass parent_class;
};

G_DEFINE_TYPE (LarkRHVoice, lark_rhvoice, LARK_TYPE_TTS_ENGINE)


static void lark_rhvoice_constructed(GObject *object)
{
	LarkRHVoice *self __attribute__((unused)) = (LarkRHVoice *)object;
	
	G_OBJECT_CLASS(lark_rhvoice_parent_class)->constructed(object);
}

static void lark_rhvoice_finalize(GObject *object)
{
	LarkRHVoice *self = (LarkRHVoice *)object;
	
	/*Delete TTS engine instance*/
	RHVoice_delete_tts_engine(self->engine);
	
	G_OBJECT_CLASS(lark_rhvoice_parent_class)->finalize(object);
}

static void lark_rhvoice_class_init(LarkRHVoiceClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	
	object_class->constructed = lark_rhvoice_constructed;
	object_class->finalize = lark_rhvoice_finalize;
}

static gint lark_rhvoice_fill_buffer(const gshort *samples, guint count, LarkRHVoice *self)
{
	g_return_val_if_fail(self != NULL, 0);
	
	g_signal_emit_by_name(G_OBJECT(self), "samples-generated", count, samples);
	
	return (gint)!*self->genstate;
}

static void lark_rhvoice_init(LarkRHVoice *self)
{
	RHVoice_callbacks callbacks;
	RHVoice_init_params init_params;
	RHVoice_voice_info *voices;
	gchar **profiles, **profvoices;
	guint voicesnum, profilesnum;
	gint p, i, v;
	GSList *voicelist;
	LarkTtsEngineVoice *voice;
	
	/*Callbacks*/
	callbacks.play_speech = (rhvoice_play_speech_func)lark_rhvoice_fill_buffer;
	callbacks.process_mark = NULL;
	callbacks.play_audio = NULL;
	callbacks.sentence_ends = NULL;
	callbacks.sentence_starts = NULL;
	callbacks.word_ends = NULL;
	callbacks.word_starts = NULL;
	
	/*Init params*/
	init_params.callbacks = callbacks;
	init_params.data_path = NULL;
	init_params.config_path = NULL;
	init_params.resource_paths = NULL;
	init_params.options = RHVoice_preload_voices;
	
	/*Create new TTS engine instance*/
	self->engine = RHVoice_new_tts_engine(&init_params);
	
	/*Get languages list*/
	voicesnum = RHVoice_get_number_of_voices(self->engine);
	voices = (RHVoice_voice_info *)RHVoice_get_voices(self->engine);
	
	profilesnum = RHVoice_get_number_of_voice_profiles(self->engine);
	profiles = (gchar **)RHVoice_get_voice_profiles(self->engine);
	
	voicelist = NULL;
	
	for (p = 0; p < profilesnum; p++) {
		/*New voice entry*/
		voice = g_new0(LarkTtsEngineVoice, 1);
		voice->name = g_strdup(profiles[p]);
		voice->languages = g_ptr_array_new_with_free_func((GDestroyNotify)g_free);
		voice->genders = g_array_new(FALSE, TRUE, sizeof(gint));
		/*Languages list*/
		profvoices = g_strsplit(profiles[p], "+", -1);
		if (profvoices != NULL) {
			for (i = 0; profvoices[i] != NULL; i++) {
				for (v = 0; v < voicesnum; v++) {
					if (g_strcmp0(profvoices[i], voices[v].name) == 0) {
						g_ptr_array_add(voice->languages, g_strdup(voices[v].language));
						g_array_append_val(voice->genders, voices[v].gender);
						break;
					}
				}
			}
			g_strfreev(profvoices);
		}
		/*Add voice entry to list*/
		voicelist = g_slist_prepend(voicelist, voice);
	}
	
	voicelist = g_slist_reverse(voicelist);
	
	g_object_set(self, "voices", voicelist, "format", LARK_RHVOICE_FORMAT, "sample-rate", LARK_RHVOICE_SAMPLERATE, "channels", LARK_RHVOICE_CHANNELS, NULL);
}

gboolean lark_rhvoice_generate_samples(LarkRHVoice *self, gchar *line, gboolean *genstate)
{
	RHVoice_message message;
	RHVoice_synth_params synthparams;
	RHVoice_capitals_mode capmode;
	RHVoice_punctuation_mode punctmode;
		
	g_return_val_if_fail(self != NULL, FALSE);
	g_return_val_if_fail(line != NULL, FALSE);
	g_return_val_if_fail(genstate != NULL, FALSE);
	
	self->genstate = genstate;
	
	switch (lark_tts_engine_get_capitals_indication(LARK_TTS_ENGINE(self))) {
		case LARK_TTS_CAPITALS_SYMBOLS_INDICATION_PITCH:
			capmode = RHVoice_capitals_pitch;
			break;
		case LARK_TTS_CAPITALS_SYMBOLS_INDICATION_VOLUME:
			capmode = RHVoice_capitals_sound;
			break;
		case LARK_TTS_CAPITALS_SYMBOLS_INDICATION_OFF:
			capmode = RHVoice_capitals_off;
			break;
		default:
			capmode = RHVoice_capitals_default;
			break;
	}
	
	switch (lark_tts_engine_get_punctuation_indication(LARK_TTS_ENGINE(self))) {
		case LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_ON:
			punctmode = RHVoice_punctuation_all;
			break;
		case LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_OFF:
			punctmode = RHVoice_punctuation_none;
			break;
		default:
			punctmode = RHVoice_punctuation_default;
			break;
	}
	
	synthparams.voice_profile = lark_tts_engine_get_voice_profile(LARK_TTS_ENGINE(self));
	synthparams.absolute_rate = lark_tts_engine_get_rate(LARK_TTS_ENGINE(self));
	synthparams.relative_rate = 1.0;
	synthparams.absolute_pitch = lark_tts_engine_get_pitch(LARK_TTS_ENGINE(self));
	synthparams.relative_pitch = 1.0;
	synthparams.absolute_volume = lark_tts_engine_get_volume(LARK_TTS_ENGINE(self));
	synthparams.relative_volume = 1.0;
	synthparams.punctuation_mode = punctmode;
	synthparams.punctuation_list = NULL;
	synthparams.capitals_mode = capmode;
	
	message = RHVoice_new_message(self->engine, line, strlen(line), RHVoice_message_text, &synthparams, self);
	
	if (message == NULL) return FALSE;
	
	RHVoice_speak(message);
	
	RHVoice_delete_message(message);
	
	return TRUE;
}

LarkRHVoice *lark_rhvoice_new(void)
{
	return g_object_new(lark_rhvoice_get_type(), NULL);
}
